(*
 Sia GAB l’insieme dei grafi orientati con nodi etichettai su A ed archi etichet-
 tati su B, con le seguenti operazioni:
	mkG: crea un grafo vuoto GAB, per dati tipi A e B;
	addN ln g: genera un grafo g’ di GAB che estende g con i nodi della lista ln 
			   non gi\`a contenuti in g. 
	addE le g: genera un grafo g’ di GAB che estende g con gli archi della lista 
			   le non gi\`a contenuti in g. Pu\`o solevare eccezioni.
	removeN n g: genera un grafo g' che differisce da g per la rimozione del nodo 
				 n e di tutto gli archi entranti ed uscenti da n. Pu\`o sol. eccez.
	removeE e g: genera un grafo g' che differisce da g per la rimozione del solo 
	             arco e. Pu\'o sollevare eccezioni.
	isNin n g: calcola vero sse n \`e un nodo di g
	isEin e g: calcola vero sse e \`e un arco di g
	nodes g: calcola la lista di tutti i nodi di g
	outN n g:  calcola la lista dei nodi uscenti da n in g. Pu\`o sollevare eccez.  
 Si definisca in OCaml un modulo signature, GRAPH, per il tipo astratto ('a,'b)gr-
 aph che esprime grafi GAB IMMUTABE con le operazioni descritte sopra.
 (a) Si mostri intestazione dell’assunta classe GRAPH, segnatura, modificatori e 
     valori calcolati dei metodi
 (b) Si mostri un modulo Graph che fornisca una ADT
 Allo scopo si assumano dati i tipi:
 		'a node per nodi etichettati su 'a;
 		('a,'b) edge per archi etichettati su 'b tra nodi di tipo 'a node.
 *)

(* questi tipi sono pubblici per semplicit\`a *) 
type 'a node = N of 'a
type ('a,'b) edge = E of 'a node * 'a node * 'b


(* questo \`e il tipo astratto richiesto *)
module type GRAPH =
 sig type ('a,'b) graph 
  val mkG: 'a -> 'b -> ('a,'b) graph
  val addN: 'a node list -> ('a,'b) graph -> ('a,'b) graph
  val addE: ('a,'b) edge list -> ('a,'b) graph -> ('a,'b) graph
  val removeN: 'a node -> ('a,'b) graph -> ('a,'b) graph
  val removeE: ('a,'b) edge -> ('a,'b) graph -> ('a,'b) graph
  val isNin: 'a node -> ('a,'b) graph -> bool
  val isEin: ('a,'b) edge -> ('a,'b) graph -> bool
  val nodes: ('a, 'b) graph -> 'a node list
  val outN: 'a node -> ('a,'b) graph -> 'a node list
  exception InvalidArgException of string
end;;
 
module Graph =
 (struct
  type ('a,'b) outer = C of 'a node * ('a,'b) edge list
  type ('a,'b) graph = G of 'a * 'b * ('a,'b) outer list
  (*
  	AF(c) = <{},{}> iff (c = G(_,_,[]))
  	AF(c) = <{n1,...,nk},{e1,...,eh}> iff ((c = G(_,_,otr)) && 
  			                               forall i\in[1..k], (node ni otr) &&
  			                               forall i\in[1..h], (edge ei otr))
  	I(c) = 	nodes(c) = duplex(nodes(c) && edges(c) = duplex(edges(c)) &&
  			(c = G(_,_,otr) && (outerCheck otr) && 			
  	        ((edge (E(_,n2,_)) otr) => (node n2 otr)) 	
  	where:
  	node, edge, nodes, edges sono le funzioni definite sotto nell'ADT.
  	outerCheck otr = let g n = fun a (E(n1,_,_) -> if (n = n1) then a else false in
  					 let check1 (C(n,ln)) = List.fold_left (g n) true ln in
  					 let g = fun a x -> a && (ceck1 x) in
  					 List.fold_left g true otr   	 	
  *)
  
(*metodi ausiliari:inizio*)
  let getE (G(labN,labE,edges)) = edges 
  let node x otr = let g (C(n,_)) = n in List.mem x (List.map g otr)
  let edge x otr = let g (C(_,el)) = el in List.mem x (List.flatten(List.map g otr)) 
  let duplex xx =
 		let isIn = (fun x -> let g = (fun a v -> (x = v)||a) in
 	                         List.fold_left g false) in 
 	    let g = (fun x a -> if (isIn x a) then a else x::a) in 
 		List.fold_right g xx []	
  let outerN (C(n,_)) = n
  let outerE (C(_,el)) = el
  let target (E(s,t,b)) = t
  let getLN(G(anull,bnull,otr)) =  anull
  let getLE(G(anull,bnull,otr)) =  bnull
  let findN n otr = (* n deve essere un nodo del grafo *)
  		let g = fun a o -> if (n = outerN o) then o else a in
  		List.fold_left g (List.hd otr) (List.tl otr)  
  		
(*metodi pubblici:inizio*)
  exception InvalidArgException of string
  let mkG anull bnull = G(anull,bnull,[])
  let addN nn g = (* rimuovere duplicati da nn *)
  		let otr = getE g in
  		let h = (fun n a -> if node n otr then a else C(n,[])::a) in
		let otr1 = (List.fold_right h nn1 otr) in 
		G(getLN g, getLE g, otr1)
  let addE ee g = g (* da definire: simile a sopra *)
  let removeN n g = let h = (fun x a -> if (outerN x = n) then a else x::a) in
  				    let otr1 = List.fold_right h (getE g) [] in
  				    let rem = (fun t elist -> 
  				    	let g = (fun x a -> if (target x = t) then a else x::a) in
  				    	List.fold_right g elist []) in
  				    let h = fun (C(k,el)) -> C(k,rem n el) in
  				    let otr2 = List.map h otr1 in
  				    G(getLN g, getLE g, otr2)
  let removeE e g =  g (* da definire: banale *)
  let isNin n (G(_,_,otr)) = node n otr
  let isEin e (G(_,_,otr)) = edge e otr
  let nodes (G(_,_,otr)) = List.map outerN otr
  let edges (G(_,_,otr)) = List.flatten (List.map outerE otr)
  let outN n (G(_,_,otr)) = 
  	        if not(node n otr) then raise(InvalidArgException "outN")
  			else let C(_,el) = findN n otr in List.map target el
  let toString g toStringN toStringE = let (G(_,_,otr)) = g in match otr with
  		|[] -> "<{},{}>"
  		|_ -> let h toString = (fun x a -> (toString x) ^", "^a) in
  		      let nn = "{" ^ (List.fold_right (h toStringN) (nodes g) "") ^ "}" in
  		      let ee = "{" ^ (List.fold_right (h toStringE) (edges g) "") ^ "}" in 
  		      "<" ^ nn ^ "," ^ ee ^ ">"	  					
end:GRAPH);;
